import json
import logging
import websockets


class BaseServer:
    name = ''

    def __init__(self, host, port, ssl=None):
        self.log = logging.getLogger(self.__class__.__name__)
        self.host = host
        self.port = port
        self.ssl = ssl

    async def start(self):
        self.log.info(f'Listening on {self.host}:{self.port}')
        if self.ssl:
            self.log.info('Authorized CAs:')
            for cert in self.ssl.get_ca_certs():
                subject = dict(x[0] for x in cert['subject'])
                self.log.info(' - ' + (','.join('='.join(x) for x in subject.items())))

        await websockets.serve(
            lambda w, _: self.handler(w),
            self.host,
            self.port,
            ssl=self.ssl,
        )


class BaseWorker:
    def __init__(self, websocket):
        self.websocket = websocket
        self.closed = False
        self.log = logging.getLogger(f'{self.__class__.__name__}[{self}]')

    def __str__(self):
        return f'{self.websocket.remote_address[0]}:{self.websocket.remote_address[1]}'

    async def start(self):
        raise NotImplementedError

    async def recv_service_message(self):
        return json.loads(await self.websocket.recv())

    async def send_service_message(self, data):
        await self.websocket.send(json.dumps(data))

    async def fatal(self, code, **kwargs):
        await self.send_service_message(
            {
                '_': 'error',
                'code': code,
                **kwargs,
            }
        )
        await self.close()

    async def wait(self):
        pass

    async def close(self):
        if not self.closed:
            self.closed = True
            self.log.info(f'Closing connection {self}')



import asyncio
import hmac
import logging
from typing import Set
import websockets

from .base import BaseServer, BaseWorker

log = logging.getLogger('server')


class GatewayWorker(BaseWorker):
    def __init__(self, server, websocket):
        self.server = server
        self.target_host = None
        self.target_port = None
        self._socket_reader = None
        self._websocket_reader = None
        self.reader = None
        self.writer = None
        super().__init__(websocket)
        self.log.info(f'New connection {self}')

    async def start(self):
        await self.send_service_message(
            {
                '_': 'hello',
                'version': 1,
                'auth_required': True,
            }
        )
        msg = await self.recv_service_message()
        if msg.get('_') != 'hello':
            return await self.fatal('expected-hello')

        if not msg.get('auth_token'):
            return await self.fatal('expected-auth-token')

        valid_tokens = list(self.server.authorized_tokens)
        if self.server.permanent_auth_token:
            valid_tokens.append(self.server.permanent_auth_token)

        for token in valid_tokens:
            if hmac.compare_digest(msg['auth_token'], token):
                if token in self.server.authorized_tokens:
                    self.server.authorized_tokens.remove(token)
                break
        else:
            return await self.fatal('incorrect-auth-token')

        await self.send_service_message({'_': 'ready'})

        msg = await self.recv_service_message()
        if msg.get('_') != 'connect':
            return await self.fatal('expected-connect')

        self.target_host = msg.get('host')
        self.target_port = msg.get('port')

        try:
            self.reader, self.writer = await asyncio.open_connection(
                self.target_host, self.target_port
            )
        except Exception as e:
            self.log.info(f'Connection failed for {self}: {e}')
            return await self.fatal('connection-failed', details=str(e))

        self.log.info(f'Connection established for {self}')
        await self.send_service_message({'_': 'connected'})

        self._socket_reader = asyncio.get_event_loop().create_task(self.socket_reader())
        self._websocket_reader = asyncio.get_event_loop().create_task(
            self.websocket_reader()
        )

    async def wait(self):
        if self._socket_reader:
            await self._socket_reader
        if self._websocket_reader:
            await self._websocket_reader

    async def close(self):
        await super().close()
        if not self.closed:
            try:
                if self.writer:
                    await self.writer.drain()
            except ConnectionResetError:
                pass
            if self.writer:
                self.writer.close()

    async def websocket_reader(self):
        while True:
            if self.closed:
                return
            try:
                data = await self.websocket.recv()
            except websockets.exceptions.ConnectionClosed:
                await self.close()
                return
            self.writer.write(data)

    async def socket_reader(self):
        while True:
            if self.closed:
                return
            try:
                await self.reader._wait_for_data('read')
            except TimeoutError:
                # TODO handle
                await self.close()
                return

            data = bytes(self.reader._buffer.copy())
            if not data:
                await self.close()
                return
            del self.reader._buffer[:]
            await self.websocket.send(data)


class GatewayServer(BaseServer):
    authorized_tokens: Set[str] = set()

    def __init__(self, *, host, port, ssl=None, auth_token=None):
        super().__init__(host, port, ssl)
        self.permanent_auth_token = auth_token

    async def start(self):
        if self.permanent_auth_token:
            log.info('Token auth enabled')

        await super().start()

    async def handler(self, websocket):
        w = GatewayWorker(self, websocket)
        await w.start()
        await w.wait()
        await w.close()



import argparse
import asyncio
import logging
import os
import ssl

from .admin_server import AdminServer
from .gateway_server import GatewayServer


async def _main():
    parser = argparse.ArgumentParser(
        formatter_class=argparse.ArgumentDefaultsHelpFormatter
    )
    parser.add_argument('--host', default='127.0.0.1', help='address to listen on')
    parser.add_argument('--port', default=9000, type=int, help='port to listen on')
    parser.add_argument(
        '--certificate', metavar='PATH', help='path to the SSL certificate. Enables SSL'
    )
    parser.add_argument('--private-key', metavar='PATH')
    parser.add_argument(
        '--ca',
        metavar='PATH',
        help='path to the CA certificate. Enables SSL client auth',
    )
    parser.add_argument(
        '--token-auth',
        action='store_true',
        help='enables token based auth using the token from the TABBY_AUTH_TOKEN env var',
    )
    parser.add_argument(
        '--admin-host',
        default='127.0.0.1',
        help='address to listen on for management requests',
    )
    parser.add_argument(
        '--admin-port',
        type=int,
        help='port to listen on for management requests',
    )
    parser.add_argument(
        '--admin-certificate',
        metavar='PATH',
        help='path to the SSL certificate for the management server',
    )
    parser.add_argument('--admin-private-key', metavar='PATH')
    parser.add_argument(
        '--admin-ca',
        metavar='PATH',
        help='path to the CA certificate for the management server',
    )
    args = parser.parse_args()

    if args.certificate and not args.private_key:
        parser.error('--private-key is required if --certificate is set')

    if args.token_auth and not os.getenv('TABBY_AUTH_TOKEN'):
        parser.error('TABBY_AUTH_TOKEN must be provided when using --token-auth')

    if args.admin_port:
        if (
            not args.admin_ca or
            not args.admin_certificate or
            not args.admin_private_key
        ):
            parser.error(
                '--admin-ca, --admin-certificate and --admin-private-key are all required when --admin-port is set'
            )

    logging.basicConfig(level=logging.INFO)

    ssl_context = None
    if args.certificate:
        ssl_context = ssl.create_default_context(
            purpose=ssl.Purpose.CLIENT_AUTH,
            cafile=os.path.realpath(args.ca) if args.ca else None,
        )
        ssl_context.load_cert_chain(
            os.path.realpath(args.certificate),
            os.path.realpath(args.private_key),
        )
        if args.ca:
            ssl_context.verify_mode = ssl.CERT_REQUIRED

    admin_ssl_context = None
    if args.admin_ca:
        admin_ssl_context = ssl.create_default_context(
            purpose=ssl.Purpose.CLIENT_AUTH,
            cafile=os.path.realpath(args.admin_ca),
        )
        admin_ssl_context.load_cert_chain(
            os.path.realpath(args.admin_certificate),
            os.path.realpath(args.admin_private_key),
        )
        admin_ssl_context.verify_mode = ssl.CERT_REQUIRED

    await GatewayServer(
        host=args.host,
        port=args.port,
        ssl=ssl_context,
        auth_token=os.getenv('TABBY_AUTH_TOKEN'),
    ).start()

    if args.admin_port:
        await AdminServer(
            host=args.admin_host,
            port=args.admin_port,
            ssl=admin_ssl_context,
        ).start()


def main():
    asyncio.get_event_loop().run_until_complete(_main())
    try:
        asyncio.get_event_loop().run_forever()
    except KeyboardInterrupt:
        os._exit(0)


if __name__ == '__main__':
    main()



import logging
from tabby_connection_gateway.gateway_server import GatewayServer
from .base import BaseServer, BaseWorker

log = logging.getLogger('server')


class AdminWorker(BaseWorker):
    async def start(self):
        await self.send_service_message(
            {
                '_': 'hello',
                'version': 1,
            }
        )
        msg = await self.recv_service_message()
        if msg.get('_') == 'authorize-client':
            token = msg.get('token')
            if not token:
                await self.fatal('expected-token')
            GatewayServer.authorized_tokens.add(token)
            self.log.info(f'Added a new auth token {token[:4]}***')
        else:
            await self.fatal('unexpected-command')


class AdminServer(BaseServer):
    async def handler(self, websocket):
        w = AdminWorker(websocket)
        await w.start()
        await w.wait()
        await w.close()



#!/usr/bin/env python
from setuptools import setup

with open('README.md') as readme_file:
    readme = readme_file.read()

requirements = ['websockets>=9.1']

setup(
    name='tabby-connection-gateway',
    version='0.2.2',
    author='Eugene Pankov',
    author_email='e@ajenti.org',
    python_requires='>=3.6',
    entry_points={
        'console_scripts': [
            'tabby-connection-gateway=tabby_connection_gateway.cli:main',
        ],
    },
    install_requires=requirements,
    license='MIT license',
    long_description=readme,
    long_description_content_type='text/markdown',
    include_package_data=True,
    keywords='tabby',
    packages=['tabby_connection_gateway'],
    url='https://github.com/eugeny/tabby-connection-gateway',
)
